# -*- coding: UTF-8 -*-
#
#   ::::::::> awsac-processjobs (08-12-15) <::::::::
#
#   by Jan-Philip Gehrcke (jgehrcke@gmail.com)
#   Universität Würzburg, Max-Planck-Institut für Physik München
# 
#   Copyright (C) 2008 Jan-Philip Gehrcke
#
#   LICENSE:                      
#   This program is free software; you can redistribute it and/or modify 
#   it under the terms of the GNU General Public License as published by 
#   the Free Software Foundation; either version 3 of the License, or 
#   (at your option) any later version. This program is distributed in
#   the hope that it will be useful, but WITHOUT ANY WARRANTY; without
#   even the implied warranty of MERCHANTABILITY or FITNESS FOR A
#   PARTICULAR PURPOSE. See the GNU General Public License for more
#   details. You should have received a copy of the GNU General Public
#   License along with this program; if not, see
#   <http://www.gnu.org/licenses/>.
#
#   read the detailled documentation at http://gehrcke.de/awsac
#
########################################################################

import sys
import os
import time
import subprocess
import optparse
import tarfile
import csv
import boto

VERSION = 'v2008-12-15'

# set the jobs' CWD (they create their own jobnumber-subfolder):
MAINWORKINGDIR = '/mnt/atlasworkarea'
# here we will look for the job shell scripts (AWSACworkingDir):
AWSACworkingDir = '/mnt/awsac'
# here the EBS volumes will be mounted
ATLASDIR = '/mnt/atlas'

def main():
    """
    lead through the script step by step
    """

    # split stdout+stderr to 'standard-stdout' and to logfile
    # (with the help of `Tee`-class)
    logfilepath = 'processjobs.log'
    log_fd = open(logfilepath,'w+')
    sys.stdout = Tee(sys.stdout, log_fd)
    sys.stderr = sys.stdout

    # parse commandline arguments and check user given files
    options = parse_args()
    files = check_usergiven_files(options)

    # start output and get information about this instance and its jobs
    timestring = time.asctime(time.localtime())
    print '\n****** awsac-processjobs.py start '+timestring+' ******\n'
    print '# parsing the sessioninfofile '+files['infofilepath']+' ...'
    session_info_dict = parse_sessioninfofile(files['infofilepath'])

    print '  Session ID: '+session_info_dict['sessID']+' ; ',
    print 'CoresPerInstance: '+session_info_dict['CoresPerInstance']
    print '# getting information about this instance from EC2...'
    thisinstance = check_instance(options.instance_id)
    thislaunchindex = thisinstance.ami_launch_index
    thisinstance_id = thisinstance.id
    thiszone = thisinstance.placement
    print ('  the ami-launch-index of this instance '
           '('+thisinstance_id+') is '+str(thislaunchindex))
    print '  the Availability Zone of this instance is '+thiszone
    print '# parsing the jobsconfigfile '+files['jobsfilepath']+' ...'
    jobinfodicts_list = parse_jobconfigfile(jobsfile=files['jobsfilepath'],
                                          launchindex=int(thislaunchindex),
                                          cpi=int(session_info_dict['CoresPerInstance']))
    print str(jobinfodicts_list)

    # handle EBS volume for this instance
    print '# prepare the creation of EBS volume(s) from '+\
          'the needed ATLAS Release snapshot(s)...'
    ebs_volumes_list = init_EBSvols(jobinfodicts_list, thisinstance_id, thiszone)

    ebsvolumes = EBSVolumes(ebs_volumes_list)

    print '# create EBS volume(s)...'
    ebsvolumes.create_vols_from_snaps()

    print '# attach EBS volume(s)...'
    ebsvolumes.attach_volumes()

    print '# mount EBS volume(s)...'
    ebsvolumes.mount_volumes()

    print '# this is the list of successfully mounted EBS - snapshots:'
    mounted_snap_ids_list = ebsvolumes.get_mounted_snap_ids()
    print '  '+str(mounted_snap_ids_list)

    print '# the following jobs now have their desired EBS running in system:'
    jobinfodicts_list = filter_jobs_by_snap_id(jobinfodicts_list, mounted_snap_ids_list)
    print str(jobinfodicts_list)

    # build joblist: EVERY planned job for THIS instance gets
    # its `Job`-object and is appended to the joblist
    print '# initialize running jobs...'
    joblist = []
    for jobinfodict in jobinfodicts_list:
        joblist.append(Job(jobinfodict))

    # create an object `jobs` of class `Jobs`.
    # Gets the joblist to contain all planned jobs. Gets session information.
    jobs = Jobs(joblist=joblist,
                session_id=session_info_dict['sessID'],
                bucketname=session_info_dict['bucketname'],
                launchindex=thislaunchindex,
                instance_id=thisinstance_id)

    # run the jobs. this method will create a subprocess for each job
    jobs.runjobs()

    # check the status of the jobs' subprocesses. do things like stdout+err saving,
    # uploading result-archives etc.. look down for details.
    # loop ends when all subprocesses have finished.
    jobs.checkjobs_loop()

    # print some results..
    jobs.verbose()

    # clean up the system
    print '# unmount EBS volume(s)...'
    ebsvolumes.unmount_volumes()

    print '# detach EBS volume(s)...'
    ebsvolumes.force_all_detach()

    print '# delete EBS volume(s)...'
    ebsvolumes.delete_all()

    # stop splitting stdout+stderr and save logfile of this script
    print '# close processjobs-logfile...'
    print '  it will then be bundled and uploaded. then EC2 will be instructed to terminate this instance'
    upload_processjobs_log(session_id=session_info_dict['sessID'],
                           bucketname=session_info_dict['bucketname'],
                           launchindex=thislaunchindex,
                           logfilehandle=log_fd,
                           logfilepath=logfilepath)

    # shut the system down
    print '# instructing EC2 to terminate my instance..'
    terminate_this_instance(thisinstance_id)


class EBSVolumes(object):
    """
    Manages a list of instances of class `EBSVolume`.

    Provides methods that operate on all `EBSVolume`-instances.
    """

    def __init__(self, ebs_volumes_list):
        self.ebs_volumes_list = ebs_volumes_list

    def create_vols_from_snaps(self):
        for ebsvolume in self.ebs_volumes_list:
            ebsvolume.create()

    def attach_volumes(self):
        for ebsvolume in self.ebs_volumes_list:
            ebsvolume.attach()

    def mount_volumes(self):
        for ebsvolume in self.ebs_volumes_list:
            ebsvolume.mount()

    def unmount_volumes(self):
        for ebsvolume in self.ebs_volumes_list:
            ebsvolume.unmount()

    def get_mounted_snap_ids(self):
        mounted_snap_ids_list = []
        for ebsvolume in self.ebs_volumes_list:
            if ebsvolume.mountsuccess:
                mounted_snap_ids_list.append(ebsvolume.snap_id)
        return mounted_snap_ids_list

    def force_all_detach(self):
        for ebsvolume in self.ebs_volumes_list:
            ebsvolume.force_detach()

    def delete_all(self):
        for ebsvolume in self.ebs_volumes_list:
            ebsvolume.delete()


class EBSVolume(object):
    """
    One object of this class represents one physical EBS Volume.
    """

    def __init__(self, snap_id, device, thisinstance_id, thiszone):
        self.snap_id = snap_id
        self.thiszone = thiszone
        self.device = device
        self.botovolumeobject = None
        self.thisinstanceId = thisinstance_id
        self.createsuccess = False
        self.mountsuccess = False
        self.attachsuccess = False
        self.detachsuccess = False

    def exists_snap_id(self):
        """
        Check if there exists an EC2 EBS snapshot with the given ID (using boto)

        @return: True or False
        """
        conn = boto.connect_ec2()
        snapshots = conn.get_all_snapshots()
        for snapshot in snapshots:
            if snapshot.id == self.snap_id:
                return True
        return False

    def create(self):
        """
        Create an EC2 EBS volume of the given snapshot ID (using boto).
        Try it for some time.
        """
        if self.exists_snap_id():
            conn = boto.connect_ec2()
            self.botovolumeobject = conn.create_volume(size=20,
                                                       zone=self.thiszone,
                                                       snapshot=self.snap_id)
            print '  instructed EC2 to create EBS from snapshot '+self.snap_id
            if self.botovolumeobject is not None:
                volumestatus = self.botovolumeobject.status
                volumeid_list = [self.botovolumeobject.id]
                for counter in range(30):
                    print '  status: '+volumestatus
                    if volumestatus == 'available':
                        self.createsuccess = True
                        break
                    if counter == 29:
                        print '  we will no longer wait. check EBS manually...'
                    time.sleep(3)
                    conn = boto.connect_ec2()
                    volumes = conn.get_all_volumes(volume_ids=volumeid_list)
                    for volume in volumes:
                        volumestatus = volume.status
        else:
            print 'processjobs.py: snapshot '+self.snap_id+' does not exist'

    def attach(self):
        """
        Attach a created EC2 EBS volume to the given device (using boto).
        Try it for some time.
        """
        if self.createsuccess:
            conn = boto.connect_ec2()
            volumestatus = conn.attach_volume(self.botovolumeobject.id,
                                              self.thisinstanceId,
                                              self.device)
            print '  instructed EC2 to attach volume '+self.botovolumeobject.id+' to '+self.device
            volumeid_list = [self.botovolumeobject.id]
            for counter in range(30):
                print '  status: '+volumestatus
                if volumestatus == 'attached':
                    self.attachsuccess = True
                    break
                if counter == 29:
                    print '  we will no longer wait. check EBS manually...'
                time.sleep(3)
                conn = boto.connect_ec2()
                volumes = conn.get_all_volumes(volume_ids=volumeid_list)
                for volume in volumes:
                    volumestatus = volume.status

    def force_detach(self):
        """
        Force detaching an attached EC2 EBS volume (using boto).
        Try it for some time.
        """
        if self.attachsuccess:
            conn = boto.connect_ec2()
            volumestatus = conn.detach_volume(volume_id=self.botovolumeobject.id,
                                              instance_id=self.thisinstanceId,
                                              force=True)
            print '  instructed EC2 to detach volume '+self.botovolumeobject.id
            volumeid_list = [self.botovolumeobject.id]
            for counter in range(30):
                print '  status: '+volumestatus
                if volumestatus == 'available':
                    self.detachsuccess = True
                    break
                if counter == 29:
                    print '  we will no longer wait. check EBS manually...'
                time.sleep(3)
                conn = boto.connect_ec2()
                volumes = conn.get_all_volumes(volume_ids=volumeid_list)
                for volume in volumes:
                    volumestatus = volume.status

    def delete(self):
        """
        Delete a detached EC2 EBS volume (using boto). Try it for some time.
        """
        if self.detachsuccess:
            conn = boto.connect_ec2()
            deletestatus = conn.delete_volume(volume_id=self.botovolumeobject.id)
            print '  instructed EC2 to delete volume '+self.botovolumeobject.id
            for counter in range(30):
                print '  status: '+str(deletestatus)
                if deletestatus == True:
                    self.deletesuccess = True
                    break
                if counter == 29:
                    print '  we will no longer wait. check EBS manually...'
                time.sleep(3)
                conn = boto.connect_ec2()
                deletestatus = conn.delete_volume(volume_id=self.botovolumeobject.id)

    def mount(self):
        """
        Mount an attached EC2 EBS volume to given device using subprocess module.
        """
        if self.attachsuccess:
            argumentlist = ['mount',self.device,ATLASDIR]
            # mount-subprocess. collect stderr and stdout in PIPE (file-like-object)
            print '  invoke mounting: subprocess.Popen() with args '+str(argumentlist)
            sp = subprocess.Popen(args=argumentlist,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT)
            returncode = sp.wait()
            output = sp.stdout.read()
            print '  mount subprocess ended. returncode: '+str(returncode)
            if output:
                print 'stdout+stderr: '+output
            if returncode == 0:
                self.mountsuccess = True


    def unmount(self):
        """
        Unmount a mounted EC2 EBS volume from given device using subprocess module.
        """
        if self.mountsuccess:
            argumentlist = ['umount',self.device]
            # umount-subprocess. collect stderr and stdout in PIPE (file-like-object)
            print '  invoke unmounting: subprocess.Popen() with args '+str(argumentlist)
            sp = subprocess.Popen(args=argumentlist,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT)
            returncode = sp.wait()
            output = sp.stdout.read()
            print '  umount subprocess ended. returncode: '+str(returncode)
            if output:
                print 'stdout+stderr: '+output


class Job(object):
    """
    An instance of class `Job` contains information about one single job.
    """
    def __init__(self, jobinfodict):
        self.shellscript = jobinfodict['shscript']
        self.jobnr = jobinfodict['jobnr']
        self.workingdir = os.path.join(MAINWORKINGDIR, self.jobnr)
        self.outfilepath = os.path.join(self.workingdir,'stdouterr_job_'+self.jobnr+'.log')
        self.starttime = None
        self.endtime = None
        self.returncode = None
        self.executiontime = None
        self.my_subprocess = None
        self.sdb_itemname = None


class Jobs(object):
    """
    Manages all jobs (instances of class `Job`) of this EC2-instance.
    
    Provides methods for running all jobs, collecting and uploading their data to S3,
    updating jobs' status in AWS SimpleDB.
    """
    def __init__(self, joblist, session_id, bucketname, launchindex, instance_id):
        self.joblist = joblist
        self.session_id = session_id
        self.bucketname = bucketname
        self.launchindex = launchindex
        self.instance_id = instance_id

    def sdb_create_jobitems(self):
        """
        Use boto to create a new jobitem for each job in the SimpleDB domain
        'session ID'. Initialize these items with some attributes.
        """
        sdb = boto.connect_sdb()
        domain = sdb.create_domain(domain_name=self.session_id)
        for job in self.joblist:
            itemname = 'job'+job.jobnr
            newitem = domain.new_item(itemname)
            newitem['jobnr'] = job.jobnr
            newitem['status'] = 'pending'
            newitem['instance_id'] = self.instance_id
            newitem['launchindex'] = self.launchindex
            newitem['shellscript'] = job.shellscript
            job.sdb_itemname = newitem.name

    def sdb_update_job(self, job, status):
        """
        Use boto to update the status of one single `job` in SimpleDB:
        (re)set the attribute 'status' and add some more information to the DB.
        """
        sdb = boto.connect_sdb()
        domain = sdb.create_domain(domain_name=self.session_id)
        item = domain.get_item(job.sdb_itemname)
        if status == 'runstart':
            item['status'] = 'running'
            item['runstarttime'] = time.strftime('%y-%m-%d %H:%M:%S',time.localtime())
        if status == 'runend':
            item['status'] = 'waitforsave'
            item['runendtime'] = time.strftime('%y-%m-%d %H:%M',time.localtime())
            item['returncode'] = job.returncode
        if status == 'savestart':
            item['status'] = 'saving'
            item['savestarttime'] = time.strftime('%y-%m-%d %H:%M:%S',time.localtime())
        if status == 'saveend':
            item['status'] = 'finished'
            item['saveendtime'] = time.strftime('%y-%m-%d %H:%M:%S',time.localtime())


    def upload_results_archive(self, jobnr):
        """
        Look for results.tar.bz2 in the workingfolder ob job with `jobnr`.
        Upload the file to S3 using boto to
        bucket/session_id/results_job_jobnr.tar.bz2
        """
        for job in self.joblist:
            if job.jobnr == jobnr:
                tarfilepath = os.path.join(job.workingdir,'results.tar.bz2')
                if os.path.exists(tarfilepath):
                    print '  found '+tarfilepath+' ('+str(os.path.getsize(tarfilepath))+' Byte)'
                    key = self.session_id+'/results_job_'+job.jobnr+'.tar.bz2'
                    S3_upload_file(file=tarfilepath,
                                   bucketname=self.bucketname,
                                   keyname=key)
                else:
                    print '  jobresultsarchive does not exist: '+tarfilepath

    # write stdout+stderr of job into logfile. compress logfile. upload archive.
    def create_upload_stdouterr_archive(self, jobnr):
        """
        Look for stdout+stderr-`outfilepath` of job with `jobnr`.
        Bundle it as bz2 and upload it to S3 using boto to
        bucket/session_id/stdouterr_job_jobnr.tar.bz2
        """
        for job in self.joblist:
            if job.jobnr == jobnr:
                if os.path.exists(job.outfilepath):
                    logarchived = False
                    try:
                        print '  outputfile of job '+job.jobnr+' found: '+job.outfilepath
                        tarfilename = 'stdouterr_job_'+job.jobnr+'.tar.bz2'
                        tarfilepath = os.path.join(job.workingdir,tarfilename)
                        tar = tarfile.open(tarfilepath, 'w:bz2')
                        tar.add(job.outfilepath,os.path.basename(job.outfilepath))
                        tar.close()
                        logarchived = True
                        print '  bundled '+job.outfilepath+' to '+tarfilepath
                    except:
                        print '  error in archiving log of job '+job.jobnr
                    if logarchived:
                        print '  planing to upload '+tarfilepath+' ('+str(os.path.getsize(tarfilepath))+' Byte)...'
                        key = self.session_id+'/'+tarfilename
                        S3_upload_file(file=tarfilepath,
                                       bucketname=self.bucketname,
                                       keyname=key)
                else:
                    print '  logfile of job '+job.jobnr+' does not exist: '+job.outfilepath

    def runjobs(self):
        """
        Create one SimpleDB item for each job
        Iterate through all `job`s (basically shellscripts):
         - create individual directory for each job /mnt/atlasworkarea/jobnr
         - set this dir as cwd when calling `/bin/sh shellscript` with
           subprocess.Popen() method.
         - stdout and stderr of a subprocess are redirected into `job.outfilepath`
        Starting subprocesses this way is non-blocking.
        Save `job`s subprocess-objects in the `my_subprocess`-attributes
        Invoke SimpleDB update for running jobs
        """
        print '  creating SimpleDB jobitems...'
        self.sdb_create_jobitems()
        for job in self.joblist:
            print '# preparing job '+job.jobnr+'...'
            print '  cwd for job: '+job.workingdir,
            try:
                os.makedirs(job.workingdir)
                print " (created)"
            except OSError:
                print " (exists)"
            argumentlist = ['/bin/sh',job.shellscript]
            # now start the subprocess. collect stderr
            # and stdout in PIPE (file-like-object)
            print '  calling subprocess.Popen() with args: '+str(argumentlist)
            job.starttime = time.time()
            outfile = open(job.outfilepath,'w')
            job.my_subprocess = subprocess.Popen(args=argumentlist,
                                                 stdout=outfile,
                                                 stderr=subprocess.STDOUT,
                                                 cwd=job.workingdir)
            print '  runstart: updating sDB item '+job.sdb_itemname
            self.sdb_update_job(job, 'runstart')

    def checkjobs_loop(self):
        """
        Wait for all subprocesses to finish.
        If one finishes, upload its produced data, save returncode, update SimpleDB
        """
        print '# wait for subprocesses to finish...'
        while True:
            time.sleep(0.5)

            # are there running processes left? if not: break loop!
            occupied = False
            for job in self.joblist:
                if job.my_subprocess is not None:
                    occupied = True
                    break
            if not occupied:
                print '# all subprocesses ended'
                break

            # check for returnstate of subprocesses...
            for job in self.joblist:
                if job.my_subprocess is not None:
                    returncode = job.my_subprocess.poll()
                    if returncode is not None:
                        job.returncode = returncode
                        job.endtime = time.time()
                        print '# subprocess for job '+job.jobnr+' ended.',
                        print ' returncode: '+str(job.returncode)
                        print '  runend: updating sDB item '+job.sdb_itemname
                        self.sdb_update_job(job, 'runend')
                        print '  savestart: updating sDB item '+job.sdb_itemname
                        self.sdb_update_job(job, 'savestart')
                        self.upload_results_archive(job.jobnr)
                        self.create_upload_stdouterr_archive(job.jobnr)
                        print '  saveend: updating sDB item '+job.sdb_itemname
                        self.sdb_update_job(job, 'saveend')
                        # mark subprocess as finished
                        job.my_subprocess = None

    def verbose(self):
        """
        Print information about all `job`s 
        """
        print '# summary:'
        for job in self.joblist:
            job.executiontime = time.strftime("%H:%M", time.gmtime(job.endtime-job.starttime))
            print ''
            outstring = 'Jobnumber: '+job.jobnr+'\n'+\
                        'starttime: '+time.strftime('%y-%m-%d %H:%M:%S',time.localtime(job.starttime))+'\n'+\
                        'endtime: '+time.strftime('%y-%m-%d %H:%M',time.localtime(job.endtime))+'\n'+\
                        'executiontime: '+job.executiontime+'\n'+\
                        'returncode: '+str(job.returncode)
            print outstring
            print '----------------------'


def terminate_this_instance(thisinstance_id):
    """
    Use boto to invoke 'TerminateInstances' API-call for this EC2-instance
    """
    conn = boto.connect_ec2()
    for counter in range(30):
        try:
            termresponse = conn.terminate_instances(instance_ids=[thisinstance_id])
            return
        except:
            print '  terminating resulted in an error-response. try again'
        time.sleep(10)
    print '  tried terminating often enough.. check it manually'


def upload_processjobs_log(session_id, bucketname, launchindex, logfilehandle, logfilepath):
    """
    Set sys.stdout and sys.stderr to originals, close logfilehandler, archive the
    logfile and upload it to S3 appending the launchindex to the filename.
    """
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__
    logfilehandle.close()
    tarfilepath = 'processjobslog_LI_'+launchindex+'.tar.bz2'
    tar = tarfile.open(tarfilepath, 'w:bz2')
    tar.add(logfilepath,os.path.basename(logfilepath))
    tar.close()
    print '  bundled '+logfilepath+' to '+tarfilepath
    S3_upload_file(file=tarfilepath,
                   bucketname=bucketname,
                   keyname=session_id+'/'+os.path.basename(tarfilepath))


def S3_upload_file(file, bucketname, keyname):
    """
    Upload given `file`(name) to S3 bucket `bucketname` with the key `keyname` 
    """
    try:
        print time.strftime('%y-%m-%d %H:%M:%S',time.localtime())+\
              ':  start upload. bucket:'+bucketname+'; key:'+keyname
        conn = boto.connect_s3()
        bucket = conn.create_bucket(bucketname)
        k = boto.s3.key.Key(bucket)
        k.key=keyname
        k.set_contents_from_filename(file)
        print time.strftime('%y-%m-%d %H:%M:%S',time.localtime())+':  finished.'
    except:
        print 'error while uploading '+file+' to bucket:'+bucketname+' ; key:'+keyname


class Tee(object):
    """
    Delivers a write()-method that writes to two filedescriptors.

    One should be standard-stdout and the other should describe a real file.
    If sys.stdout is replaced with an instance of this `Tee`-class and sys.stderr is
    set to sys.stdout, all stdout+stderr of the script is collected to console and
    to file at the same time.
    """
    def __init__(self, stdout, file):
        self.stdout = stdout
        self.file = file

    def write(self, data):
        self.stdout.write(data)
        try:
          self.file.write(data)
          self.file.flush()
        except:
          pass


def filter_jobs_by_snap_id(jobinfodicts_list, mounted_snap_ids_list):
    """
    Check if the snap_ids in the jobinfodicts are in the list of mounted snap_ids

    @return: A list of jobinfodicts that only contains jobs with mounted snap_ids
    """
    return [dict for dict in jobinfodicts_list if dict['snap_id'] in mounted_snap_ids_list]



def parse_jobconfigfile(jobsfile, launchindex, cpi):
    """
    Determine the jobs for this instance from jobsconfigfile (given by commandline).

    jobsconfigfile contains information for all EC2-instances in this session.
    The rows must have following format: snap_id;JobScriptname;Njobs
    Read in the jobsconfigfile and save data in jobsdatadicts_list.

    Grab the jobs for THIS EC2-instance from jobsdatadicts_list.
    The algorithm for determining these jobs from launchindex and cores per instance
    is explained below.

    @return: A list of dictionaries. The list has the following format:
          [{'shscript':'/path/shellskriptA.sh','snap_id':'snap-a232329','jobnr':'3'},
           {'shscript':'/path/shellskriptB.sh','snap_id':'snap-a232329','jobnr':'4'}]
    """
    jobsdatadicts_list = []
    try:
        filehandler = open(jobsfile)
        csvreader = csv.reader(filehandler, delimiter=';')
        for jobsdata in csvreader:
            jobsdatadict = {}
            jobsdatadict['snap_id']=jobsdata[0]
            jobsdatadict['shellscript']=jobsdata[1]
            jobsdatadict['Njobs']=jobsdata[2]
            jobsdatadicts_list.append(jobsdatadict)
        print "# jobsdatadicts_list read from "+jobsfile
    except:
        sys.exit("could not read in data from jobsfile "+jobsfile)

    # myjobinfodicts_list must be of following format:
    # [{'shscript':'/path/shellskriptA.sh','snap_id':'snap-a232329','jobnr':'3'},
    #  {'shscript':'/path/shellskriptB.sh','snap_id':'snap-a232329','jobnr':'4'}]

    myjobinfodicts_list = []
    # Determine my job numbers from launch index and Cores Per Instance
    my_first_jobnumber = launchindex*cpi+1
    my_last_jobnumber = launchindex*cpi+cpi
    my_jobnumbers = range(my_first_jobnumber, my_last_jobnumber+1)
    print "# my job numbers are: "+str(my_jobnumbers)

    # now filter my jobs out of jobs.cfg
    checkjobs_begin_number = 1
    for jobsdatadict in jobsdatadicts_list:
        checkjobs_end_number_plus_1 = checkjobs_begin_number +\
                                      int(jobsdatadict['Njobs'])
        jobnumbers_of_current_jobsdata = range(checkjobs_begin_number,
                                               checkjobs_end_number_plus_1)
        # is one my job numbers in the list of job numbers of current job?
        for myjobnumber in my_jobnumbers:
              if myjobnumber in jobnumbers_of_current_jobsdata:
                  # found a jobsdatadict containing a job for me->save jobsdata
                  myjobinfodict = {}
                  myjobinfodict['shscript'] = AWSACworkingDir+'/'+\
                                              jobsdatadict['shellscript']
                  myjobinfodict['snap_id'] = jobsdatadict['snap_id']
                  myjobinfodict['jobnr'] = str(myjobnumber)
                  myjobinfodicts_list.append(myjobinfodict)
        # in the first run of this loop the jobs 1-Njobs_first (e.g. 1-5) have
        # to be checked. this is the way I do:
        # - checkjobs_begin_number = 1
        # - checkjobs_end_number_plus_1=checkjobs_begin_number+Njobs_first = 6
        # range(1,6) returns (1,2,3,4,5) - this is the list we want!
        # lets declare the first 5 jobs as processed after the first loop run.
        # in the second run of the loop we want range(6,6+Njobs_second)
        # procuding the list (6,7,...,6+Njobs_second-1).
        # so checkjobs_begin_number must be increased by Njobs_first: 
        checkjobs_begin_number += int(jobsdatadict['Njobs'])
    return myjobinfodicts_list


def parse_sessioninfofile(sessioninfofile):
    """
    Parse the sessioninfofile (given by commandline).

    The file contains information about the session this EC2-instance belongs to.
    The one and only row in this file must have the following format:
    sessID;bucket;sessionarchivefilename;CoresPerInstance

    Read in the data and save it in a dictionary session_info_dict.

    @return: dictionary session_info_dict containing session information.
    """
    session_info_dict = {}
    try:
        data = open(sessioninfofile).readline()
        data = data.split(';')
        session_info_dict['sessID'] = ''.join(data[0].split())
        session_info_dict['bucketname'] = ''.join(data[1].split())
        session_info_dict['archivename'] = ''.join(data[2].split())
        session_info_dict['CoresPerInstance'] = str(int(''.join(data[3].split())))
    except:
        sys.exit(('error in sessioninfofile. expecting `sessID;bucket;'
                  'sessionarchivefilename;CoresPerInstance` in '+sessioninfofile))
    return session_info_dict


def check_instance(instance_id):
    """
    Check if instance_id (given by commandline) belongs to a running EC2-instance
    using boto. If this EC2-instance exists, return a boto `Instance`-Class object.
    If not, it makes no sense to go on with the script.

    @return: boto `Instance`-Class object representing THIS EC2-instance
    """
    conn = boto.connect_ec2()
    reservations = conn.get_all_instances()
    for reservation in reservations:
        for instance in reservation.instances:
            if instance.id == instance_id:
                return instance
    sys.exit('InstanceID given by commandline was not found '+\
             'in currently running instances.')


def check_usergiven_files(options):
    """
    Check if the files given by commandline exist. `checkfile`-fkt will raise errors.

    @return: A dictionary containing the absolute paths of the files.
    """
    infofile = None
    jobsfile = None
    infofilepath = checkfile(options.infofile)
    jobsfilepath = checkfile(options.jobsfile)
    return {'infofilepath': infofilepath,
            'jobsfilepath': jobsfilepath}


def parse_args():
    """
    Parse commandlineoptions using the optparse module and check them for the
    logical consistence.

    @return: optparse `options`-object containing the commandlineoptions
    """
    # the following part configures the OptionParser...
    version = 'AWSACtools: processjobs'
    description = ("job system example")
    parser = optparse.OptionParser(version=version,description=description)
    parser.add_option('--sessioninfofile', dest='infofile',
                      help='job session info file (with userdata string)')
    parser.add_option('--jobsfile', dest='jobsfile',
                      help='jobs.cfg (containing jobinfo)')
    parser.add_option('--instanceID', dest='instance_id',
                      help='ID of THIS instance')

    # now read in the given arguments (from sys.argv by default)
    (options, args) = parser.parse_args()

    # now check the logical consistence...
    if (options.infofile is None) or \
       (options.jobsfile is None) or \
       (options.instance_id is None):
        parser.error('--sessioninfofile FILE and --jobsfile FILE and '+\
                     '--instanceID XXX are needed!')

    # everything okay until here? seems so.. return the given options.
    return options


def checkfile(file):
    """
    Check if a given file exists and really is a file (e.g. not a directory)
    In errorcase the script is stopped.

    @return: the absolute path of the file
    """
    if not os.path.exists(file):
        sys.exit(file+' does not exist')
    if not os.path.isfile(file):
        sys.exit(file+' must be a file')
    return os.path.abspath(file)



def init_EBSvols(jobinfodicts_list, thisinstance_id, thiszone):
    """
    Creates list of correctly initialized EBSVolume-classobjects.

    Detect the different snap-IDs for the instances jobs. Ideally its only one!
    Store them in different_snap_ids.
    Iterate through different_snap_ids:
    - choose a device path from a list of possible devices that does not exist in
      filesystem
    - delete this device path from the list of possible devices
    - Construct a new instance of `EBSVolume` with this chosen device and other
      information (availabilityzone, instanceid, snapshotid)
    - append this new instance of `EBSVolume` to the list ebs_volumes_list

    Verbose the ebs_volumes_list

    @return: ebs_volumes_list
    """
    # at first lets collect all different snapshot IDs
    different_snap_ids = []
    for jobinfodict in jobinfodicts_list:
        if jobinfodict['snap_id'] not in different_snap_ids:
            different_snap_ids.append(jobinfodict['snap_id'])
    print '  detected following different snap_ids: '+str(different_snap_ids)
    
    # this is the list of devices we can populate with EBS
    devices = ['/dev/sdh1','/dev/sdh2','/dev/sdh3',
               '/dev/sdh4','/dev/sdh5','/dev/sdh6',
               '/dev/sdh7','/dev/sdh8','/dev/sdh9']

    # now create one object of EBSVolume for each individual snap_id.
    # assign an individual device to each individual EBSVolume.
    # after this loop we have a list ebs_volumes_list that contains information
    # about individual EBS volumes to create from `snap_id` and to mount to `device`.
    ebs_volumes_list = []
    for snap_id in different_snap_ids:
        if devices:
            for possibledevice in devices:
                if not os.path.exists(possibledevice):
                    device = possibledevice
                    break
            devices.remove(device)
        else:
            sys.exit('processjobs.py: more than 9 EBS per instance are currently not supported!')
        ebs_volumes_list.append(EBSVolume(snap_id=snap_id, device=device,
                                          thisinstance_id=thisinstance_id,
                                          thiszone=thiszone))
    print '  planing to assign snap_id(s) to following device(s):'
    for ebsvolume in ebs_volumes_list:
        print '  '+ebsvolume.snap_id+'->'+ebsvolume.device
    return ebs_volumes_list


if __name__ == '__main__':
    main()
